#ifndef LATTICEPROTEIN_H_
#define LATTICEPROTEIN_H_


#include "BackboneStructure3D.hh"
#include "BioMolecule.hh"
#include "LatticeModel.hh"
#include "ContactEnergyFunction.hh"

namespace biu
{
		/*! A LatticeProtein represents a backbone protein representation
		 * in an energy landscape
		 * whereby all backbone edges are placed on integer lattice points
		 * and consecutive edges are neighbored in the lattice.
		 * 
		 * Such a structure is valid, if and only if it is selfavoiding.
		 * 
		 * The energy is based on an ContactEnergyFunction object. A contact
		 * is formed, if two structure positions are neighbored and not
		 * consecutive in sequence.
		 * 
		 * @author Martin Mann, Sebastian Will, Andreas Richter
		 */
	class LatticeProtein : public BioMolecule, public BackboneStructure3D
	{
		// needed to apply a move on the relative move sequence
	friend class PivotMoveSet;
	friend class PullMoveSet;

	protected:
			/*! the lattice model this structure belongs to */
		LatticeModel const *	lattice;		
		
			/*! the contact energy function */
		ContactEnergyFunction const *	energy;	
		
			/*! the indexed sequence of the lattice protein */
		Sequence*				sequence;		
		
			/*! the structure information coded as indexed relative moves*/
		MoveSequence*			relMoves;
		
			//! the structure in coordinates
		mutable IPointVec*		points;	
		
			//! whether or not the point coordinates have to be updated
		mutable bool			pointsToUpdate;
		
			//! access to the coordinates, ensuring that these are up to date
		IPointVec*		getPoints() const;
		
			//! removes point coordinates if relMoves are changed
		void 			clearPoints();
		
	public:
			/*! Constructs a new lattice protein object. 
			 * @param lattice		the lattice model the protein structure is 
			 * 						basing on
			 * @param energy 		the energy function inclusive the alphabet
			 * 						for the sequence
			 * @param seq			the sequence string
			 * @param moveString 	a move string representation of the 
			 * 						structure
			 * @param isAbsoluteMove == true if moveString is an absolute move 
			 * 						string or == false if it is a relative 
			 * 						move string
			 * */
		LatticeProtein(	const LatticeModel* lattice, 
						const ContactEnergyFunction* energy, 
						const std::string& seq, 
						const std::string& moveString, 
						const bool isAbsoluteMove);
						
			/*! Copy constructor */
		LatticeProtein(const LatticeProtein& latPr);
		
		virtual ~LatticeProtein();
		
		LatticeProtein&	operator= (const LatticeProtein& latProt2);		
		bool	 		operator== (const LatticeProtein& latProt2) const;
		bool	 		operator!= (const LatticeProtein& latProt2) const;

		
			/*! Returns the absolute move string representation of the lattice
			 * protein structure. */
		std::string getAbsoluteMoveString() const;
		
			/*! Returns the absolute move sequence of the lattice
			 * protein structure. */
		MoveSequence getAbsoluteMoves() const;
		
			/*! Returns the relative move string representation of the lattice
			 * protein structure. */
		std::string getRelativeMoveString() const;
		
			/*! Returns the relative move sequence of the lattice
			 * protein structure. */
		MoveSequence getRelativeMoves() const;

	// abstract functions (BackboneStructure3D)
		
		virtual DPointVec	get3Ddata() const ;
		
		virtual double		getDRMSD(const BackboneStructure3D& other) const;

	// abstract functions (BioMolecule)
		
		virtual Sequence	getSequence() const;
		
			//! Returns the relative move representation of the lattice
			//! protein structure. 
		virtual Structure	getStructure() const;
		
			/*! Returns the contact energy of the lattice protein structure.
			 * Consecutive positions are not taken into account. 
			 * Each contact is count only once.
			 */
		virtual double		getEnergy() const;
		
			/*! Returns whether or not the lattice protein is connected 
			 * and self avoiding.
			 */
		virtual bool		isValid() const;
		
			/*! Returns a combination of absolute move string representation 
			 * and sequence information of this lattice protein.
			 * 
			 * ABSOLUTEMOVESTRING(SEQUENCE) */
		virtual std::string	getStringRepresentation() const;	
		
	// additional functions 
	
			//! Returns a constant reference to the relative move representation 
			//! of the lattice protein structure. 
		virtual const Structure&	getStructureRef() const;
		
			//! Tests whether or not the structure is self avoiding.
		bool	isSelfAvoiding() const;		
		
			//! Tests whether or not consecutive structure elements 
			//! are neighbored in the lattice. 
		bool	isConnected() const;	
		
			//! Returns the lattice model this lattice protein is basing on
		const LatticeModel* const getLatticeModel() const {
			return lattice;
		}
		
			//! Returns the energy function this lattice protein is basing on
		const ContactEnergyFunction* const getContactEnergyFunction() const {
			return energy;
		}
		
		bool
		writeXYZfile(std::string & filename) const;
		
	};

} // namespace biu

#include "LatticeProtein.icc"

#endif /*LATTICEPROTEIN_H_*/
