#ifndef PULLMOVESET_HH_
#define PULLMOVESET_HH_

#include "LatticeMoveSet.hh"
#include <vector>

namespace biu
{

/*!
 * This class handles pull moves on LatticeProtein objects.
 * 
 * @author Daniel Maticzka
 */
class PullMoveSet : public LatticeMoveSet
{
public: 
	class PullMoveDecoder 
	{
	private:
		IPointVec neighDirections;
		std::vector< std::vector< IPointVec > > stdPulls;
		std::vector< IPointVec > endPulls;
		int directionNumber;
		size_t stdPullNumber;
		size_t endPullNumber;
		
	public:
		PullMoveDecoder(const LatticeModel* lattice);
		void lookupMove(const size_t& proteinLength, 
				size_t& moveIndex, bool& pullFront, 
				bool& stdPull, size_t& movePosition) const;
		IPointVec lookupCL(const IntPoint& rootpoint, 
				const IntPoint& fixpoint, 
				const size_t& moveIndex) const;
		IPointVec lookupEndMove(const IntPoint& rootpoint,
				const size_t& moveIndex) const;
		size_t getMoveNumber(const size_t proteinLength) const; 
	};
	
private:
	class RelativeInt
	{
	private:

		int value;
		
		// basic functions
		const int add(const int& a, const int& b) const { return a + b; }
		const int sub(const int& a, const int& b) const { return a - b; }
		const bool largerEq(const int& a, const int& b) const { return a >= b; }
		const bool smallerEq(const int& a, const int& b) const { return a <= b; }
		
		// relative functions for algorithm
		const int (PullMoveSet::RelativeInt::*relAdd)(const int&, const int&) const;
		const int (PullMoveSet::RelativeInt::*relSub)(const int&, const int&) const;
		const bool (PullMoveSet::RelativeInt::*relLEQ)(const int&, const int&) const;
		const bool (PullMoveSet::RelativeInt::*relSEQ)(const int&, const int&) const;

	public:
		
		RelativeInt(const int _value, const bool direction) {
			value = _value;
			relAdd = direction ? &biu::PullMoveSet::RelativeInt::add 
								: &biu::PullMoveSet::RelativeInt::sub;
			relSub = direction ? &PullMoveSet::RelativeInt::sub 
					 			: &PullMoveSet::RelativeInt::add;
			relLEQ = direction ? &PullMoveSet::RelativeInt::largerEq 
								: &PullMoveSet::RelativeInt::smallerEq;
			relSEQ = direction ? &PullMoveSet::RelativeInt::smallerEq 
								: &PullMoveSet::RelativeInt::largerEq;
		}
		
		const bool
		operator >= (const int& y) const {
			return (*this.*relLEQ)(value, y);
		}
		
		const bool
		operator <= (const int& y) const {
			return (*this.*relSEQ)(value, y);
		}	
		
		const int
		operator+ (const int& y) const {
			return (*this.*relAdd)(value, y);
		}
		
		const int
		operator- (const int& y) const {
			return (*this.*relSub)(value, y);
		}
		
		void
		operator --(int) {
			value = (*this.*relSub)(value,1);
		}
		
		void
		operator ++(int) {
			value = (*this.*relAdd)(value,1);
		}
		
		const int
		getValue() const { return value; };
	};
	
protected:
	struct UndoRecord
	{
		//! pointer to the last object that was changed
		LatticeProtein*	lastChangedObject;
		//! true if protein has been changed by last move
		bool hasChanged;
		//! pull direction at last pull
		bool pullFront;
		//! position at last pull
		size_t position;
		//! last position pulled at last pull
		size_t lastPullPosition;
		//! first overwritten position at last pull
		IntPoint lostPos0;
		//! second overwritten position at last pull
		IntPoint lostPos1;
	} undoRec;
	
public:
	PullMoveSet::PullMoveDecoder* decoder;

	/*!
	 * Constructs a PullMoveSet object operating on the LatticeModel lattice.
	 */
	PullMoveSet(const LatticeModel* lattice); 

	virtual ~PullMoveSet();

	/*! Applies a pull move to the given LatticeProtein on the 
	 * specified position and returns a new LatticeProtein object 
	 * as result.
	 * 
	 * @param todo the initial state
	 * @param position  the structure position where to apply the move
						which has to be larger than 0.
	 * @param moveIndex the move index (has to be less than the return 
	 * 					value of getMoveNumber())
	 * @return a new LatticeProtein object neighbored to the todo object
	 *
	 *  In case of invalid moves, the behaviour
	 *  of the program is not specified.
	 */
	virtual LatticeProtein* applyMove(const LatticeProtein* const todo, 
			size_t position, size_t moveIndex) ;

	/*! Applies a pull move to the given LatticeProtein on the 
	 * specified position INPLACE and returns the modified object
	 * as result.
	 *
	 * @param todo the initial state to modify
	 * @param position  the structure position where to apply the move
								which has to be larger than 0.
	 * @param moveIndex the move index (has to be less than the return 
	 * 					value of getMoveNumber())
	 * @return the modified LatticeProtein 
	 *
	 *  In case of invalid moves, the behaviour
	 *  of the program is not specified.
	 */
	virtual LatticeProtein* applyMoveInPlace(LatticeProtein* todo, size_t position,
			size_t moveIndex) ;

	/*!
	 * Returns number of moves that can be applied to the given lattice.
	 */
	virtual unsigned int getMoveNumber() const;
	size_t getMoveNumber ( const LatticeProtein* const lp ) const; 

	/*!
	 * Undo of last pull move performed on LatticeProtein toUndo. Undo
	 * is performed in place!
	 */
	virtual LatticeProtein* undoLastMove(LatticeProtein* toUndo);
};

} // end namespace biu

#endif /*PULLMOVESET_HH_*/
